#  Copyright (c) 2020 ETH Zurich

"""
Module for the MonteCarlo class.

The MonteCarlo class implements all random-number based methods to simulate the motion
of electrons. The simulation time-step is determined with the null collision technique
(calculate_max_coll_freq creates a lookup table for the choice of the
trial collision frequency, determine_timestep calculates the time-step based on the
current maximum electron energy and acceleration). The collision processes are randomly
chosen based on the collision frequency of each process. The scattering angles are
randomly chosen with either an isotropic or an anisotropic model.
"""

# Import Packages
import numpy as np
import numpy.linalg as linalg
import scipy.constants as csts
import scipy.interpolate
from typing import Tuple

# Import modules
import pyMETHES.utils as utils
from pyMETHES.__about__ import __version__
from pyMETHES.config import Config
from pyMETHES.gas_mixture import GasMixture

np.seterr(all='raise')


class MonteCarlo:
    """
    Class implementing all random-number based simulation methods

    Attributes:
        config (Config): configuration of the simulation
        trial_coll_freq (float): trial collision frequency for the null collision
            technique
        max_coll_freq (interp1d): cumulative maximum of the collision frequency as a
            function of the electron energy
        max_coll_period (ndarray): array inversely proportional to the cumulative
            maximum of the collision frequency
        max_coll_period_squared (ndarray): array inversely proportional to the square
            of the cumulative maximum of the collision frequency
        collision_by_electron (ndarray): array of collision index for each electron
            (starting at 0), an index equal to the number of cross sections indicates
            a null collision
    """

    version = f"pyMETHES version {__version__}\n"

    def __init__(self, cfg: Config):
        """
        Instantiates the MonteCarlo class.

        Args:
            cfg (Config): configuration of the simulation
        """

        self.config = cfg

        # trial collision frequency
        self.trial_coll_freq = None
        # data for the calculation of the trial collision frequency
        self.max_coll_freq = None
        self.max_coll_period = None
        self.max_coll_period_squared = None

        # vector of which collision which electron undergoes
        self.collision_by_electron = None

    def calculate_max_coll_freq(self, gas_mixture: GasMixture):
        """
        Calculates the maximum collision frequency in the given gas mixture.

        Args:
            gas_mixture (GasMixture): gas mixture
        """

        gas_density = self.config.gas_number_density
        velocity = utils.velocity_from_energy(gas_mixture.energy_vector)
        freq = gas_density * velocity * gas_mixture.total_cross_section
        # replace zero value to avoid divide by zero:
        freq[0] = freq[1]
        # take cumulative maximum of the collision frequency (make it monotonous)
        freq = np.maximum.accumulate(freq)
        self.max_coll_freq = scipy.interpolate.interp1d(gas_mixture.energy_vector,
                                                        freq,
                                                        kind='linear')
        self.max_coll_period = \
            0.5 * csts.electron_mass / csts.elementary_charge / freq
        self.max_coll_period_squared = \
            0.5 * csts.electron_mass / csts.elementary_charge / freq ** 2

    def determine_timestep(self, max_velocity: float, max_acceleration: float) -> float:
        """
        Determine the duration of the next time-step in the simulation with the
        null-collision technique.

        Args:
            max_velocity (float): current maximum electron velocity
            max_acceleration (float): current maximum electron acceleration

        Returns: time-step duration (s)
        """

        # Draw random number
        s = - np.log(np.random.random())
        # Calculate energy after step (worst case: max_velocity + max_acceleration * dt)
        max_energy = utils.energy_from_velocity(max_velocity)
        freq = self.max_coll_freq.y
        de = 2 * max_velocity * max_acceleration * s * self.max_coll_period \
            + max_acceleration ** 2 * s ** 2 * self.max_coll_period_squared
        # Clip energy after step to maximum energy
        e_end = np.clip(max_energy + de, a_min=None,
                        a_max=self.config.max_cross_section_energy)
        # Find trial collision frequency that fulfills the condition
        ok_frequencies = freq[self.max_coll_freq(e_end) <= freq]
        # Take the first (= smallest) frequency that fulfills the condition
        self.trial_coll_freq = 1.01 * ok_frequencies[0]
        # Calculate the time-step
        dt = s/self.trial_coll_freq
        return dt

    def determine_collisions(self, gas_mixture: GasMixture,
                             velocity_norm: np.ndarray,
                             energy: np.ndarray) -> None:
        """
        Calculates the collision frequencies for all electrons with all cross sections,
        and chooses a collision type via a random number.

        Args:
            gas_mixture (GasMixture): cross section data
            velocity_norm (ndarray): norm of the velocity of each electron
            energy (ndarray): energy of each electron
        """

        # calculate the coll frequency for each cross section and each electron
        collision_matrix = np.array([
            self.config.gas_number_density * interp(energy) * velocity_norm
            / self.trial_coll_freq for interp in gas_mixture.cross_sections])

        rand_vector = np.random.rand(1, energy.size)
        rand_matrix = np.repeat(rand_vector, collision_matrix.shape[0], axis=0)
        coll_matrix_cum = np.cumsum(collision_matrix, axis=0)
        self.collision_by_electron = np.sum(rand_matrix > coll_matrix_cum, axis=0)

    def perform_collisions(
            self, gas_mixture: GasMixture, position: np.ndarray, velocity: np.ndarray,
            energy: np.ndarray) -> Tuple[np.ndarray, np.ndarray, int, int, int]:
        """
        Calculates the electrons positions (created/removed) and velocities (scattered)
        after the collisions listed in the collision_by_electron array.

        Args:
            gas_mixture (GasMixture): cross section data
            position (ndarray): coordinates (x,y,z) of each electron (m)
            velocity (ndarray): velocity of each electron in (x,y,z) directions (m.s-1)
            energy (ndarray): energy of each electron (eV)

        Returns: the new position and velocity of electrons, the number of collisions,
            cations and anions produced
        """

        null_collisions = \
            self.collision_by_electron >= gas_mixture.number_of_cross_sections

        # set aside electrons which did not collide (null collision)
        null_coll_p = position[:, null_collisions]
        null_coll_v = velocity[:, null_collisions]

        is_collision = np.logical_not(null_collisions)
        collisions = self.collision_by_electron[is_collision]
        n_coll = collisions.size
        pos = position[:, is_collision]
        vel = velocity[:, is_collision]
        ener = energy[is_collision]

        # count attached electrons
        attachment_collisions = gas_mixture.is_attachment[collisions]
        n_att = np.count_nonzero(attachment_collisions)

        # remove attached electrons by excluding them
        is_collision = np.logical_not(attachment_collisions)
        collisions = collisions[is_collision]
        pos = pos[:, is_collision]
        vel = vel[:, is_collision]
        ener = ener[is_collision]

        # calculate the new velocity direction after scattering
        v_scattered_hat, cos_chi = MonteCarlo.unit_scattered_velocity(
            ener, vel, self.config.isotropic_scattering)

        # Calculate the losses and the remaining energy
        mass_ratios = gas_mixture.mass_ratios[collisions]
        ionization_collisions = gas_mixture.is_ionization[collisions]
        thresholds = gas_mixture.thresholds[collisions]
        sharing = np.ones(thresholds.shape)
        sharing[ionization_collisions] *= self.config.energy_sharing_factor
        losses = thresholds + ener * mass_ratios * (1 - cos_chi)
        energy_after_coll = np.maximum(ener - losses, 0)

        # Calculate the new velocity
        v_scattered = \
            v_scattered_hat * utils.velocity_from_energy(
                energy_after_coll * sharing)

        # Add electrons produced by ionization
        new_e_pos = pos[:, ionization_collisions]
        new_e_energy = energy_after_coll[ionization_collisions] \
            * (1 - self.config.energy_sharing_factor)
        new_e_velocity = \
            MonteCarlo.unit_scattered_velocity(ener[ionization_collisions],
                                               vel[:, ionization_collisions], True)[0] \
            * utils.velocity_from_energy(new_e_energy)
        n_ion = new_e_energy.size

        pos_end = np.hstack([null_coll_p, pos, new_e_pos])
        vel_end = np.hstack([null_coll_v, v_scattered, new_e_velocity])

        if self.config.conserve:
            ne_ini = position.shape[1]
            ne_end = pos_end.shape[1]
            if ne_end > ne_ini:
                # delete some random electrons to compensate for the increased number
                indexes_to_keep = np.random.permutation(ne_end)[:ne_ini]
                pos_end = pos_end[:, indexes_to_keep]
                vel_end = vel_end[:, indexes_to_keep]
            elif ne_end < ne_ini:
                # duplicate some random electrons to compensate for the decreased number
                n_duplicates = ne_ini - ne_end
                # stack if required, to draw enough duplicates
                n_stacks = int(np.ceil(n_duplicates/ne_end))
                indexes_to_duplicate = np.concatenate(
                    [np.random.permutation(ne_end) for _ in range(n_stacks)]
                )[:n_duplicates]
                pos_end = np.hstack([pos_end, pos_end[:, indexes_to_duplicate]])
                vel_end = np.hstack([vel_end, vel_end[:, indexes_to_duplicate]])

        return pos_end, vel_end, n_coll, n_ion, n_att

    @staticmethod
    def unit_scattered_velocity(energy: np.ndarray, velocity: np.ndarray,
                                iso: bool) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculates the new direction of the velocity vector after scattering.

        Args:
            energy (ndarray): energy of each electron (eV)
            velocity (ndarray): velocity of each electron in (x,y,z) directions before
                the collision (m.s-1)
            iso (bool): isotropic scattering (True) or anisotropic scattering (False)

        Returns: normed velocities after collisions, cosine of polar scattering angle
        """

        cos_chi, sin_chi, cos_phi, sin_phi = MonteCarlo.scattering_angles(energy, iso)

        # Normalize the velocity
        v_hat = velocity / linalg.norm(velocity, axis=0, keepdims=True)

        # unit vector in x direction
        e_x = np.zeros_like(velocity)
        e_x[0, :] = 1

        # calculate theta (angle between v_hat and e_x)
        theta = np.arccos(v_hat[0, :])
        sin_theta = np.sin(theta)

        # calculate the direction of the new velocity, see Vahedi 1995 equation (11)
        v_new_dir = \
            v_hat * cos_chi \
            + np.cross(v_hat, e_x, axis=0) * sin_chi * sin_phi / sin_theta \
            + np.cross(v_hat, np.cross(e_x, v_hat, axis=0), axis=0) \
            * sin_chi * cos_phi / sin_theta

        return v_new_dir, cos_chi

    @staticmethod
    def scattering_angles(energy: np.ndarray, iso: bool
                          ) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray):
        """
        Generates values for the polar (chi) and azimuthal (phi)
        isotropic or anisotropic scattering angles according to Vahedi (1995).

        Args:
            energy: array of electron energies
            iso: isotropic scattering or not

        Returns: 4 arrays cos(chi), sin(chi), cos(phi), sin(phi)
        """

        # choose scattering angle phi
        phi = 2 * np.pi * np.random.random(energy.size)
        sin_phi = np.sin(phi)
        cos_phi = np.cos(phi)

        # choose scattering angle chi
        if iso:
            cos_chi = 1 - 2 * np.random.random(energy.size)
        else:
            cos_chi = (2 + energy
                       - 2 * (1 + energy) ** np.random.random(energy.size)) / energy

        sin_chi = np.sqrt(1 - cos_chi ** 2)

        return cos_chi, sin_chi, cos_phi, sin_phi



#  Copyright (c) 2020 ETH Zurich

"""
Module for the GasMixture class, which aggregates data of different gases.

The GasMixture uses the cross_section module to read cross section data. It stores
linear interpolations of each cross section, as well as the mass ratio,
energy threshold and type of cross section, which are needed for the simulation.
The proportions of the gas mixture should sum up to 1. A common energy_vector for all
cross sections is calculated, as well as the total_cross_section.
"""

# Import Packages
import numpy as np
from lxcat_data_parser import CrossSectionTypes as CST
import scipy.interpolate

# Import modules
from pyMETHES.cross_section import InterpolatedCrossSectionSet

np.seterr(all='raise')


class GasMixture:
    """
    Class representing a gas mixture.

    Attributes:
        cross_sections (ndarray): array of cross section interpolations (from all gases)
        types (ndarray): type of each cross section of each species
        thresholds (ndarray): threshold of each cross section of each species
        mass_ratios (ndarray): mass ratios (repeated for each cross section)
        is_attachment (ndarray): boolean mask for ATTACHMENT cross sections
        is_ionization (ndarray): boolean mask for IONIZATION cross sections
        energy_vector (ndarray): common energy vector, containing all distinct energies
            of all cross sections
        total_cross_section (ndarray): sum of all cross sections of all species,
            linearly interpolated at energy_vector
        number_of_cross_sections (int): total number of cross sections (from all gases)
    """

    def __init__(self, gas_formulas: list, path_to_cross_section_files: list,
                 proportions: list, max_cross_section_energy: float):
        """
        Instantiates a GasMixture.

        Args:
            gas_formulas (list): list of the chemical formulas of the species
            path_to_cross_section_files (list): path to the cross section data of each
                species
            proportions (list): proportion of each species in the mixture
            gases (list): InterpolatedCrossSectionSet of each species
            max_cross_section_energy (float): maximum cross section energy (eV)
            number_of_cross_sections (int): total number of cross sections across all
                species

        Raises:
            ValueError: If the proportions do not sum up to 1
        """

        # check if the sum of proportions is 1
        if not np.isclose(sum(proportions), 1):
            raise ValueError("The sum of the proportions of the gases in the GasMixture"
                             "must be 1, but it is %1.2f" % (sum(proportions)))

        # Create a list with a Gas Object for each gas of the mixture
        gases = [InterpolatedCrossSectionSet(max_cross_section_energy, *args)
                 for args in zip(path_to_cross_section_files, gas_formulas)]

        self.cross_sections = np.array([
            scipy.interpolate.interp1d(x.data['energy'].to_numpy(),
                                       x.data['cross section'].to_numpy() * p,
                                       kind='linear')
            for p, gas in zip(proportions, gases) for x in gas.cross_sections
        ])

        cross_sections = [x for gas in gases for x in gas.cross_sections]
        self.thresholds = np.asarray([x.threshold for x in cross_sections])
        self.mass_ratios = np.asarray([x.mass_ratio for x in cross_sections])
        self.types = [x.type for x in cross_sections]
        types_array = np.asarray(self.types)
        self.is_attachment = types_array == CST.ATTACHMENT
        self.is_ionization = types_array == CST.IONIZATION

        # combine the energy vectors into a total energy vector
        all_energies = np.concatenate([x.data['energy'] for x in cross_sections])
        self.energy_vector = np.unique(all_energies)

        # calculate the total cross section
        self.total_cross_section = \
            np.sum([interp(self.energy_vector) for interp in self.cross_sections],
                   axis=0)

    @property
    def number_of_cross_sections(self) -> int:
        """
        Returns: Total number of cross sections, all gases considered.
        """

        return self.cross_sections.size



#  Copyright (c) 2020 ETH Zurich

"""
Module for the Config class of the simulation.
"""

# Import Packages
import json5
import json
from typing import Union
import scipy.constants as csts

num = Union[int, float]


class Config:
    """
    Configuration class for the Monte-Carlo simulation.

    The configuration can be loaded from, or saved to, json or json5 files.
    Alternatively, it can be provided as, or exported to, a (nested) dictionary.
    The gas number density is not a configuration parameter, but a
    cached property of the Config class, which is computed from the pressure and
    temperature.

    Attributes:
        gases (list): sum formulae of gases
        paths_to_cross_section_files (list): paths to the cross section files in
            txt format
        fractions (list): proportions of the gases in the gas mixture
        max_cross_section_energy (float): maximum cross section energy (eV)

        output_directory (str): path to the output directory

        EN (float): E/N ratio in (Td)
        _pressure (float): gas pressure in Pa
        _temperature (float): gas temperature in K
        _gas_number_density (float): gas number density in m-3

        num_e_initial (int): initial number of electrons
        initial_pos_electrons (list): initial position [x, y, z] of the electrons'
            center of mass
        initial_std_electrons (list): initial broadening of gaussian distributed
            electrons in x, y and z direction

        num_energy_bins (int): number of energy bins to group the electrons for the
            energy distribution
        energy_sharing_factor (float): energy sharing factor for ionization collisions
        isotropic_scattering (bool): scattering: isotropic (true), non-isotropic
            according to Vahedi et al. (false)
        conserve (bool): conservation of the number of electrons
        num_e_max (int): maximum allowed electron number (when it is reached, the
            number of electrons is then conserved until simulation ends)

        w_tol (float): tolerance on the flux drift velocity. simulation ends
            when w_err/w < w_tol
        DN_tol (float): tolerance on the flux diffusion coefficient. simulation ends
            when DN_err/w < DN_tol
        num_col_max (int): maximum number of collisions during the simulation,
            simulation ends when it is reached
    """

    def __init__(self, config: Union[str, dict]):
        """
        Instantiate the config.

        Args:
            config (str, dict): path to a json or json5 config file, or dictionary.
        """

        if isinstance(config, str):
            if config.endswith('.json5'):
                with open(config, "r") as json_file:
                    config = json5.load(json_file)
            elif config.endswith('.json'):
                with open(config, "r") as json_file:
                    config = json.load(json_file)
            else:
                raise ValueError(f"Configuration file '{config}' has invalid extension."
                                 " Extensions '.json' or '.json5' are expected.")

        # gases
        input_gases = config['input_gases']
        self.paths_to_cross_section_files: (str, list) = \
            input_gases['paths_to_cross_section_files']
        self.gases: (str, list) = input_gases['gases']
        self.fractions: (float, list) = input_gases['fractions']
        self.max_cross_section_energy: num = input_gases['max_cross_section_energy']

        # output
        output = config['output']
        self.output_directory: str = output['output_directory']
        self.base_name: str = output['base_name']
        self.save_simulation_pickle: bool = output['save_simulation_pickle']
        self.save_temporal_evolution: bool = output['save_temporal_evolution']
        self.save_swarm_parameters: bool = output['save_swarm_parameters']
        self.save_energy_distribution: bool = output['save_energy_distribution']

        # physical conditions
        physical_conditions = config['physical_conditions']
        self.EN: num = physical_conditions['EN']
        self._pressure: num = physical_conditions['pressure']
        self._temperature: num = physical_conditions['temperature']
        self._gas_number_density = None

        # initial state
        initial_state = config['initial_state']
        self.num_e_initial: int = int(initial_state['num_e_initial'])
        self.initial_pos_electrons: list = initial_state['initial_pos_electrons']
        self.initial_std_electrons: list = initial_state['initial_std_electrons']

        # simulation settings
        simulation = config['simulation_settings']
        self.num_energy_bins: int = simulation['num_energy_bins']
        self.energy_sharing_factor: num = simulation['energy_sharing_factor']
        self.isotropic_scattering: bool = simulation['isotropic_scattering']
        self.conserve: bool = simulation['conserve']
        self.num_e_max: int = int(simulation['num_e_max'])

        # end conditions
        end_conditions = config['end_conditions']
        self.w_tol: float = end_conditions['w_tol']
        self.DN_tol: float = end_conditions['DN_tol']
        self.num_col_max: int = int(end_conditions['num_col_max'])

    @property
    def gas_number_density(self) -> float:
        if self._gas_number_density is None:
            self._gas_number_density = \
                self.pressure / (csts.Boltzmann * self.temperature)
        return self._gas_number_density

    @property
    def pressure(self) -> float:
        return self._pressure

    @pressure.setter
    def pressure(self, value: float):
        """
        Pressure setter. If a new value is set, resets the cache for the gas
        number density.

        :param value: pressure in Pascal
        """
        self._pressure = value
        self._gas_number_density = None

    @property
    def temperature(self) -> float:
        return self._temperature

    @temperature.setter
    def temperature(self, value: float) -> None:
        """
        Temperature setter. If a new value is set, resets the cache for the gas
        number density.

        :param value: temperature in Kelvin
        """
        self._temperature = value
        self._gas_number_density = None

    def to_dict(self) -> dict:
        """
        Returns the current configuration as a dictionary.

        Returns: dict of configuration
        """

        return {
            'input_gases': {
                'gases': self.gases,
                'paths_to_cross_section_files': self.paths_to_cross_section_files,
                'fractions': self.fractions,
                'max_cross_section_energy': self.max_cross_section_energy,
            },
            'output': {
                'output_directory': self.output_directory,
                'base_name': self.base_name,
                'save_simulation_pickle': self.save_simulation_pickle,
                'save_temporal_evolution': self.save_temporal_evolution,
                'save_swarm_parameters': self.save_swarm_parameters,
                'save_energy_distribution': self.save_energy_distribution,
            },
            'physical_conditions': {
                'EN': self.EN,
                'pressure': self.pressure,
                'temperature': self.temperature,
            },
            'initial_state': {
                'num_e_initial': self.num_e_initial,
                'initial_pos_electrons': self.initial_pos_electrons,
                'initial_std_electrons': self.initial_std_electrons,
            },
            'simulation_settings': {
                'num_energy_bins': self.num_energy_bins,
                'energy_sharing_factor': self.energy_sharing_factor,
                'isotropic_scattering': self.isotropic_scattering,
                'conserve': self.conserve,
                'num_e_max': self.num_e_max,
            },
            'end_conditions': {
                'w_tol': self.w_tol,
                'DN_tol': self.DN_tol,
                'num_col_max': self.num_col_max,
            }
        }

    def save_json5(self, path: str = 'config.json5') -> None:
        """
        Saves the current configuration to a json5 file.

        Args:
            path (str): path including the file name and extension,
                example: 'data/config.json5'
        """

        with open(path, "w") as config_file:
            json5.dump(self.to_dict(), config_file, indent=2)

    def save_json(self, path: str = 'config.json') -> None:
        """
        Saves the current configuration to a json file.

        Args:
            path (str): path including the file name and extension,
                example: 'data/config.json'
        """

        with open(path, "w") as config_file:
            json.dump(self.to_dict(), config_file, indent=2)



#  Copyright (c) 2020 ETH Zurich

"""
Module for the FluxData and BulkData classes.

Both classes are based on the TransportData abstract class, and calculate the drift
velocity and diffusion coefficient of electrons based on the temporal evolution of their
position and velocity. The BulkData class calculates the bulk transport parameters,
which describe the transport of the center-of-mass of the swarm (time derivative of
spatial averages). The FluxData class calculates the flux transport parameters
(spatial averages of time derivatives).
"""

# Import Packages
from abc import ABC
import numpy as np
from scipy import stats

# Import modules
from pyMETHES.temporal_evolution import TimeSeries

np.seterr(all='raise')


class TransportData(ABC):
    """
    Stores the drift velocity and diffusion coefficient.

    Attributes:
        w (ndarray): drift velocity along x, y and z directions (m.s-1)
        w_err (ndarray): drift velocity error along x, y and z directions (m.s-1)
        DN (ndarray): diffusion coeff. along x, y and z directions (m-1.s-1)
        DN_err (ndarray): diffusion coeff. error along x, y and z directions (m-1.s-1)
    """

    def __init__(self, gas_number_density: float):
        self.gas_number_density = gas_number_density
        self.w = np.empty(3) + np.nan
        self.w_err = np.empty(3) + np.nan
        self.DN = np.empty(3) + np.nan
        self.DN_err = np.empty(3) + np.nan


class BulkData(TransportData):
    """
    Stores the bulk drift velocity and bulk diffusion coefficient.

    Attributes:
        w (ndarray): bulk drift velocity along x, y and z directions (m.s-1)
        w_err (ndarray): bulk drift velocity error along x, y and z directions (m.s-1)
        DN (ndarray): bulk diffusion coeff. along x, y and z directions (m-1.s-1)
        DN_err (ndarray): bulk diffusion coeff. error along x, y and z directions
            (m-1.s-1)
    """

    def calculate_data(self, time_series: TimeSeries) -> None:
        """
        Calculates the bulk drift velocity and bulk diffusion coefficient values
        (time derivative of spatial averages).

        Args:
            time_series (TimeSeries): temporal evolution of some quantities
        """

        x = time_series.time[time_series.ind_equ:]
        # for x,y,z directions
        for i in range(3):
            # drift coefficients
            y = time_series.mean_position[time_series.ind_equ:, i]
            slope, _, _, _, std_err = stats.linregress(x, y)
            self.w[i] = slope
            self.w_err[i] = 1.96 * std_err

            # diffusion coefficients
            y = time_series.var_position[time_series.ind_equ:, i]
            slope, _, _, _, std_err = stats.linregress(x, y)
            self.DN[i] = 0.5 * self.gas_number_density * slope
            self.DN_err[i] = 0.5 * self.gas_number_density * 1.96 * std_err


class FluxData(TransportData):
    """
    Flux drift velocity and flux diffusion coefficient.

    Attributes:
        w (ndarray): flux drift velocity along x, y and z directions (m.s-1)
        w_err (ndarray): flux drift velocity error along x, y and z directions (m.s-1)
        DN (ndarray): flux diffusion coeff. along x, y and z directions (m-1.s-1)
        DN_err (ndarray): flux diffusion coeff. error along x, y and z directions
            (m-1.s-1)
    """

    def __init__(self, gas_number_density: float):
        super().__init__(gas_number_density)

    def calculate_data(self, time_series: TimeSeries) -> None:
        """
        Calculates the flux drift velocity and the flux diffusion coefficient
        (spatial averages of time derivatives).

        Args:
            time_series (TimeSeries): temporal evolution of some quantities
        """

        steps_since_sst = time_series.time.size - 1 - time_series.ind_equ

        # calculate the drift velocity
        self.w = np.mean(time_series.mean_velocity[time_series.ind_equ:, :], axis=0)

        # calculate the diffusion coefficient
        y = time_series.mean_velocity_moment \
            - time_series.mean_position * time_series.mean_velocity
        self.DN = self.gas_number_density * np.mean(y[time_series.ind_equ:, :], axis=0)

        # At least 10 steps are required to estimate the errors
        if steps_since_sst > 9:
            # calculate the drift velocity error
            self.w_err = \
                np.std(time_series.mean_velocity[time_series.ind_equ:, :], axis=0) \
                / np.sqrt(steps_since_sst)

            # calculate the diffusion coefficient error
            y = time_series.mean_velocity_moment \
                - time_series.mean_position * time_series.mean_velocity
            self.DN_err = self.gas_number_density \
                * np.std(y[time_series.ind_equ:, :], axis=0) / np.sqrt(steps_since_sst)



#  Copyright (c) 2020 ETH Zurich

"""
Module for the InstantaneousEnergyDistribution and TimeAveragedEnergyDistribution
classes.

Both classed inherit from the abstract class EnergyDistribution. The
InstantaneousEnergyDistribution class can be used to calculate the eedf and eepf
at a given point in time. The TimeAveragedEnergyDistribution class allows for
averaging the eedf and eepf over time. This is done by updating an histogram of
electron energies at every time-step.
"""

# Import Packages
import numpy as np
from abc import ABC

np.seterr(all='raise')


class EnergyDistribution(ABC):
    """
    Abstract base class for electron energy distributions.

    Attributes:
        energy_bins (ndarray): bins to construct the histogram of electron energies
        energy_bin_centers (ndarray): centers of the energy bins (length - 1)
        eedf (ndarray): electron energy distribution function (eV-1)
        eepf (ndarray): electron energy probability function (eV-3/2)
    """

    def __init__(self):
        self.energy_bins = None
        self.energy_bin_centers = None
        self.eedf = None
        self.eepf = None

    def calculate_bin_centers(self) -> None:
        """
        Calculate the center values of the energy bins.
        """

        self.energy_bin_centers = 0.5 * (self.energy_bins[1:] + self.energy_bins[:-1])


class InstantaneousEnergyDistribution(EnergyDistribution):
    """
    Stores the energy distribution of electrons.
    """

    def calculate_distribution(self, energies, energy_bins=None) -> None:
        """
        Calculates the eedf and eepf associated to the input electron energies.

        Args:
            energies (ndarray): array of electron energies
            energy_bins (ndarray): (optional) energy bins for calculating the eedf
                and eepf. If None, the energy bins are automatically determined with the
                Freedman Diaconis Estimator.
        """

        self.eepf = None
        if energy_bins is None:
            (self.eedf, self.energy_bins) = np.histogram(energies, bins='fd',
                                                         density=True)
        else:
            self.energy_bins = energy_bins
            self.eedf = np.histogram(energies, bins=energy_bins, density=True)[0]
        self.calculate_bin_centers()
        if 0 not in self.energy_bin_centers:
            self.eepf = self.eedf / np.sqrt(self.energy_bin_centers)


class TimeAveragedEnergyDistribution(EnergyDistribution):
    """
    Stores the energy distribution of electrons (averaged over time after swarm
    equilibration).

    Attributes:
        cumulative_energy_histogram (ndarray): energy histogram cumulated over every
            time-step since steady-state
    """

    def __init__(self):
        super().__init__()
        self.cumulative_energy_histogram = 0
        self.energy_mean = None
        self.energy_mean_err = None

    def generate_bins(self, num_bins: int, max_energy: float) -> None:
        """
        Generate fixed energy bins for averaging the eedf over time. The bins are
        spaced linearly and range from 0 to 150% of the maximum electron energy at the
        step where the method is called.

        Args:
            num_bins (int): number of energy bins
            max_energy (float): maximum electron energy (eV)
        """

        self.energy_bins = np.linspace(0, 1.5 * max_energy, num_bins)
        self.calculate_bin_centers()

    def collect_histogram(self, energies) -> None:
        """
        Updates the current eedf and eepf, using the electron energies given as input.
        The energy bins should be already defined and are re-used. This method should
        be called at every time-step.

        Args:
            energies (ndarray): array of electron energies
        """

        self.cumulative_energy_histogram += np.histogram(energies,
                                                         bins=self.energy_bins)[0]

    def calculate_distribution(self, mean_energies: np.ndarray) -> None:
        """
        Calculates the eedf and eepf using the cumulative energy histogram.
        Calculates the mean electron energy averaged over time.

        Args:
            mean_energies (ndarray): mean energies at every time-step after
                swarm equilibration
        """

        integral = sum(self.cumulative_energy_histogram * np.diff(self.energy_bins))
        self.eedf = self.cumulative_energy_histogram
        if integral != 0:
            self.eedf = self.eedf / integral
        if 0 not in self.energy_bin_centers:
            self.eepf = self.eedf / np.sqrt(self.energy_bin_centers)

        # calculate the mean energy
        self.energy_mean = np.mean(mean_energies)

        # At least 10 steps are required to estimate the errors
        if mean_energies.size > 9:
            # calculate the mean energy error
            self.energy_mean_err = np.std(mean_energies) / np.sqrt(mean_energies.size)



#  Copyright (c) 2020 ETH Zurich

"""
Module for the TimeSeries class.
"""

# Import Packages
import numpy as np
import pandas as pd

# Import modules
from pyMETHES.electrons import Electrons


class TimeSeries:
    """
    Temporal evolution of some quantities during the simulation.

    The TimeSeries class stores the temporal evolution of some quantities in arrays of
    ever-increasing length. The TimeSeries can be exported to a pandas DataFrame for
    further processing.

    Attributes:
        ind_equ (int): index of equilibration time
        time (ndarray): simulated time
        num_collisions (ndarray): cumulative number of collisions
        num_electrons (ndarray): number of electrons
        num_cations (ndarray): cumulative number of cations
        num_anions (ndarray): cumulative number of anions
        mean_energy (ndarray): mean energy of electrons
        mean_position (ndarray): mean position of electrons
        var_position (ndarray): variance of electrons positions
        mean_velocity (ndarray): mean velocity of electrons
        mean_velocity_moment (ndarray): mean velocity moment of electrons
    """

    def __init__(self, electrons: Electrons):
        """
        Instantiates a TimeSeries.

        Args:
            electrons (Electrons): electron related data
        """

        # number of time steps until equilibration
        self.ind_equ = None

        self.time = np.array([0])
        self.num_collisions = np.array([0])
        self.num_electrons = np.array([electrons.num_e])
        self.num_cations = np.array([0])
        self.num_anions = np.array([0])
        self.mean_energy = np.array([electrons.mean_energy])
        self.mean_position = electrons.mean_position
        self.var_position = electrons.var_position
        self.mean_velocity = electrons.mean_velocity
        self.mean_velocity_moment = electrons.mean_velocity_moment

    def append_data(self, electrons: Electrons, dt: float, n_coll: int, n_cations: int,
                    n_anions: int) -> None:
        """
        Appends latest data to the arrays storing the temporal evolution of some
        quantities.

        Args:
            electrons (Electrons): current data related to the electrons
            dt (float): duration of the current time-step (s)
            n_coll (int): number of collisions during the current time-step
            n_cations (int): number of cations produced during the current time-step
            n_anions (int): number of anions produced during the current time-step
        """

        self.time = np.append(self.time, self.time[-1] + dt)
        self.num_collisions = np.append(self.num_collisions,
                                        self.num_collisions[-1] + n_coll)
        self.num_electrons = np.append(self.num_electrons, electrons.num_e)
        self.num_cations = np.append(self.num_cations,
                                     self.num_cations[-1] + n_cations)
        self.num_anions = np.append(self.num_anions,
                                    self.num_anions[-1] + n_anions)
        self.mean_energy = np.append(self.mean_energy, electrons.mean_energy)
        self.mean_position = np.vstack([self.mean_position, electrons.mean_position])
        self.var_position = np.vstack([self.var_position, electrons.var_position])
        self.mean_velocity = np.vstack([self.mean_velocity, electrons.mean_velocity])
        self.mean_velocity_moment = np.vstack([self.mean_velocity_moment,
                                               electrons.mean_velocity_moment])

    def to_dataframe(self) -> pd.DataFrame:
        """
        Creates a pandas DataFrame with the data containe in the TimeSeries and
        returns it.

        Returns: pandas DataFrame containing the TimeSeries data.
        """

        n = self.time.size
        num_coll = np.diff(self.num_collisions, prepend=0)
        num_null_coll = self.num_electrons - num_coll

        df = pd.DataFrame(data=np.concatenate([
            self.time.reshape((n, 1)),
            num_coll.reshape((n, 1)),
            num_null_coll.reshape((n, 1)),
            self.num_electrons.reshape((n, 1)),
            self.num_cations.reshape((n, 1)),
            self.num_anions.reshape((n, 1)),
            self.mean_energy.reshape((n, 1)),
            self.mean_position,
            self.var_position,
            self.mean_velocity,
            self.mean_velocity_moment
        ], axis=1), columns=[
            'time',
            'num_collisions',
            'num_null_collisions',
            'num_electrons',
            'num_cations',
            'num_anions',
            'mean_energy',
            'mean_position_x',
            'mean_position_y',
            'mean_position_z',
            'var_position_x',
            'var_position_y',
            'var_position_z',
            'mean_velocity_x',
            'mean_velocity_y',
            'mean_velocity_z',
            'mean_velocity_moment_x',
            'mean_velocity_moment_y',
            'mean_velocity_moment_z',
        ])
        df.attrs = {'index_equilibration': self.ind_equ}
        return df



#  Copyright (c) 2020 ETH Zurich

__title__ = "pyMETHES"
__summary__ = "Monte Carlo simulation of electron transport in low temperature plasmas"
__url__ = "https://gitlab.ethz.ch/hvl_priv/pymethes/"
__version__ = "0.1.0"
__author__ = "Alise Chachereau, Kerry Jansen, Markus Niese"
__email__ = "alisec@ethz.ch"
__license__ = "GNU General Public License v3 (GPLv3)"
__copyright__ = "Copyright (c) 2020 ETH Zurich"



#  Copyright (c) 2020 ETH Zurich

"""
Module for importing cross section data, based on the 'lxcat_data_parser' package.

Data points are added to each cross section at zero energy and at max_energy (which
is user-defined). The classes 'CrossSection' and 'CrossSectionSet' of the
lxcat_data_parser are extended to include 'interp1d' linear interpolations of each
cross section.
"""

# Import Packages
import warnings
import pandas as pd
import molmass
import numpy as np
import scipy.interpolate
import scipy.constants as csts
import matplotlib.pyplot as plt
from lxcat_data_parser import CrossSectionTypes as CST
from lxcat_data_parser import CrossSection, CrossSectionSet, CrossSectionReadingError
from typing import Union

np.seterr(all='raise')


class InterpolatedCrossSection(CrossSection):
    """
    Extension of the CrossSection class to add a linear interpolation.

    Attributes:
        type (str, CrossSectionType): type of collision
        species (str): chemical formula of the species, example: N2
        mass_ratio (float): ratio of electron mass to molecular mass
        threshold (float): energy threshold (eV) of the cross section
        database (str): name of the database
        data (DataFrame): pandas DataFrame with columns "energy" and "cross section"
        info: optional additional information on the cross section given via kwargs
        interpolation(interp1d): linear interpolation of the cross section
    """

    def __init__(self, cross_section_type: Union[str, CST], species: str,
                 data: pd.DataFrame, mass_ratio: float,
                 threshold: float, **kwargs):
        """
        Instantiate an InterpolatedCrossSection.

        Args:
            cross_section_type (str, CrossSectionType): type of collision
            species (str): chemical formula of the species, example: N2
            data (DataFrame): pandas DataFrame with columns "energy" and "cross section"
            mass_ratio (float): ratio of electron mass to molecular mass
            threshold (float): energy threshold (eV) of the cross section
        """

        super().__init__(cross_section_type, species, data, mass_ratio,
                         threshold, **kwargs)
        self.interpolation = scipy.interpolate.interp1d(
            self.data['energy'].to_numpy(), self.data['cross section'].to_numpy(),
            kind='linear')


class InterpolatedCrossSectionSet(CrossSectionSet):
    """
    Extension of the CrossSectionSet class to use InterpolatedCrossSections.

    Attributes:
        species (str): chemical formula of the species, example: N2
        database (str): name of the database
        cross_sections (list): list of CrossSection instances
    """

    def __init__(self, max_cross_section_energy: float, input_file: str,
                 imposed_species: str, imposed_database: str = None):
        """
        Instantiate an InterpolatedCrossSectionSet.

        Args:
            max_cross_section_energy (float): maximum cross section energy (eV), which
                is appended to the end of the cross section data
            input_file (str): path to the cross section data
            imposed_species (str): chemical formula of the species, example: N2
            imposed_database (str): (optional) name of the database

        Raises:
            CrossSectionReadingError: if the provided file does not contain a valid set.
        """

        super().__init__(input_file, imposed_species, imposed_database)
        # find the mass ratio value attached to the ELASTIC or EFFECTIVE cross section
        x = [x for x in self.cross_sections if x.type in (CST.EFFECTIVE, CST.ELASTIC)]
        if not x:
            raise CrossSectionReadingError(
                "The cross section set should contain exactly one EFFECTIVE "
                "CrossSection or one ELASTIC CrossSection. None was found.")
        elif len(x) > 1:
            raise CrossSectionReadingError(
                "The cross section set should contain exactly one EFFECTIVE "
                "CrossSection or one ELASTIC CrossSection. Several were found.")
        mass_ratio = x[0].mass_ratio

        # check the mass ratio using the molmass package
        try:
            molmass_mass_ratio = csts.electron_mass \
                / (molmass.Formula(self.species).mass / (1000 * csts.Avogadro))
            if not np.isclose(mass_ratio, molmass_mass_ratio):
                warnings.warn(
                    "Incorrect mass ratio."
                    f" The mass ratio {mass_ratio} read from the file "
                    f"'{input_file}' for the species {imposed_species} does not match "
                    f"the value {molmass_mass_ratio:.4} calculated with the molmass "
                    f"package. You may consider fixing the input file."
                )
        except molmass.FormulaError:
            warnings.warn(
                "Could not check the input mass ratio using the molmass package."
                f" {imposed_species} is not a valid chemical formula."
            )

        # add a factor of 2 inside the mass ratio for future energy loss calculations
        mass_ratio *= 2

        for x in self.cross_sections:
            if x.mass_ratio is None:
                x.mass_ratio = 0
            else:
                x.mass_ratio = mass_ratio
            # set threshold to 0 when not defined
            if x.threshold is None:
                x.threshold = 0
            # add data points (if needed) at zero energy and at max energy
            if x.data['energy'].iat[0] > 0:
                zero_energy_line = pd.DataFrame(
                    {'energy': [0],
                     'cross section': [x.data['cross section'].iat[0]]})
                x.data = pd.concat([zero_energy_line, x.data], ignore_index=True)
            if x.data['energy'].iat[-1] < max_cross_section_energy:
                final_energy_line = pd.DataFrame(
                    {'energy': [max_cross_section_energy],
                     'cross section': [x.data['cross section'].iat[-1]]})
                x.data = x.data.append(final_energy_line, ignore_index=True)
            elif x.data['energy'].iat[-1] > max_cross_section_energy:
                interp = scipy.interpolate.interp1d(
                    x.data['energy'], x.data['cross section'], kind='linear')
                x.data = x.data.loc[x.data['energy'] < max_cross_section_energy]
                final_energy_line = pd.DataFrame(
                    {'energy': [max_cross_section_energy],
                     'cross section': [interp(max_cross_section_energy)]})
                x.data = x.data.append(final_energy_line, ignore_index=True)

        # replace CrossSections by InterpolatedCrossSection
        self.cross_sections = [
            InterpolatedCrossSection(
                x.type, x.species, x.data, x.mass_ratio, x.threshold, info=x.info
            ) for x in self.cross_sections
        ]

        # replace if needed the effective cross section by an elastic one
        self.effective_to_elastic()

    def effective_to_elastic(self) -> None:
        """
        If effective cross section is given instead of elastic, calculate elastic cross
        section and replace the effective cross section.
        """

        # check if there is an effective cross section
        x_eff = [x for x in self.cross_sections if x.type == CST.EFFECTIVE]
        if x_eff:

            # combine the energy vectors into a total energy vector
            all_energies = np.concatenate([x.data['energy']
                                           for x in self.cross_sections])
            energy_vector = np.unique(all_energies)

            # calculate the elastic cross section to replace the effective cross section
            x_ela = x_eff[0]
            x_ela.type = CST.ELASTIC
            x_ela.data = pd.DataFrame({
                'energy': energy_vector,
                'cross section': x_ela.interpolation(energy_vector)
            })

            # subtract all the excitation and ionization cross sections
            for x in self.cross_sections:
                if x.type in [CST.EXCITATION, CST.IONIZATION]:
                    x_ela.data['cross section'] -= x.interpolation(x_ela.data['energy'])
            # update the linear interpolation
            x_ela.interpolation = scipy.interpolate.interp1d(
                x_ela.data['energy'].to_numpy(), x_ela.data['cross section'].to_numpy(),
                kind='linear')

    def plot(self, block=True) -> None:
        """
        Plot the cross section data.

        Args:
            block (bool): block execution or not when showing plot
        """

        fig = plt.figure()
        ax = fig.add_subplot(111)

        colors = ['black', 'does_not_occur', 'grey', 'blue', 'red']
        labels = []
        for x in self.cross_sections:
            if x.type not in labels:
                labels.append(x.type)
                ax.loglog(x.data['energy'], x.data['cross section'],
                          color=colors[x.type.value], label=x.type.name)
            else:
                ax.loglog(x.data['energy'], x.data['cross section'],
                          color=colors[x.type.value])

        # some cosmetics
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5),
                   title='type of cross section')
        plt.title('Cross Section Set of ' + self.species)
        plt.xlabel('energy [eV]')
        plt.ylabel('cross section [m^2]')

        # to make space for the legend on the right
        plt.subplots_adjust(right=0.72)
        plt.show(block=block)



#  Copyright (c) 2020 ETH Zurich

"""
Module for the Electrons class.

The electrons class stores the position, velocity and acceleration of electrons
in (3,n)-sized ndarrays. The class provides numerous class properties to compute
electrons-related data such as the mean electron energy using cached class properties.
"""

# Import Packages
import numpy as np
import numpy.linalg as linalg
import matplotlib.pyplot as plt

# Import modules
import pyMETHES.utils as utils
from pyMETHES.energy_distribution import InstantaneousEnergyDistribution

np.seterr(all='raise')


class Electrons:
    """
    Stores information on a electron swarm in motion.

    Attributes:
        position (ndarray): xyz-coordinates of each electron, dim=(num_e,3)
        velocity (ndarray): xyz-velocities of each electron, dim=(num_e,3)
        acceleration (ndarray): xyz-accelerations of each electron, dim=(num_e,3)
    """

    def __init__(self, num_e_initial: int, init_pos: list,
                 initial_std: list, e_field: float):
        """
        Instantiates an electron swarm.

        Args:
            num_e_initial (int): Initial number of electrons
            init_pos (list): initial xyz position (m) of electron swarm
            initial_std (list): initial xyz std (m) of electron swarm
            e_field (float): electric field strength (V/m) along z-direction
        """

        self.position = np.zeros([3, num_e_initial])
        for i in range(3):
            self.position[i, :] = \
                init_pos[i] + np.random.normal(loc=0, scale=initial_std[i],
                                               size=num_e_initial)

        self.velocity = np.zeros([3, num_e_initial])
        self.acceleration = None
        self.accelerate(e_field)

        # cached properties
        self._mean_position = None
        self._var_position = None
        self._mean_velocity = None
        self._mean_velocity_moment = None
        self._energy = None
        self._mean_energy = None
        self._std_energy = None
        self._max_energy = None
        self._position_norm = None
        self._velocity_norm = None
        self._max_velocity_norm = None
        self._acceleration_norm = None
        self._max_acceleration_norm = None
        self._energy_distribution = InstantaneousEnergyDistribution()

    def accelerate(self, e_field: float) -> None:
        """
        Calculates the electrons acceleration in the electric field.

        Args:
            e_field (float): electric field strength (V/m) along z direction
        """

        self.acceleration = np.zeros([3, self.num_e])
        self.acceleration[2, :] = \
            utils.acceleration_from_electric_field(e_field) * np.ones(self.num_e)

    def free_flight(self, duration: float) -> None:
        """
        Update the attributes of Electrons after a free-flight in the electric field.

        Args:
            duration (float): duration (s) of the free flight
        """

        self.position += self.velocity * duration \
            + 0.5 * self.acceleration * duration ** 2
        self.velocity += self.acceleration * duration
        self.reset_cache()

    def apply_scatter(self, pos: np.ndarray, vel: np.ndarray, e_field: float) -> None:
        """
        Updates the attributes of Electrons after scattering by the gas.

        Args:
            pos (ndarray): new positions of the electrons (attachment and ionization)
            vel (ndarray): new velocities of the electrons (scattering)
            e_field (float): electric field strength (V/m) along z direction
        """

        self.position = pos
        self.velocity = vel
        self.accelerate(e_field)
        self.reset_cache()

    @property
    def num_e(self) -> int:
        return self.position.shape[1]

    @property
    def mean_position(self) -> np.ndarray:
        if self._mean_position is None:
            self._mean_position = np.mean(self.position, axis=1, keepdims=True).T
        return self._mean_position

    @property
    def var_position(self) -> np.ndarray:
        if self._var_position is None:
            self._var_position = np.var(self.position, axis=1, keepdims=True).T
        return self._var_position

    @property
    def mean_velocity(self) -> np.ndarray:
        if self._mean_velocity is None:
            self._mean_velocity = np.mean(self.velocity, axis=1, keepdims=True).T
        return self._mean_velocity

    @property
    def mean_velocity_moment(self) -> np.ndarray:
        if self._mean_velocity_moment is None:
            self._mean_velocity_moment = \
                np.mean(self.position * self.velocity, axis=1, keepdims=True).T
        return self._mean_velocity_moment

    @property
    def velocity_norm(self) -> np.ndarray:
        if self._velocity_norm is None:
            self._velocity_norm = linalg.norm(self.velocity, axis=0)
        return self._velocity_norm

    @property
    def max_velocity_norm(self) -> float:
        if self._max_velocity_norm is None:
            self._max_velocity_norm = np.max(self.velocity_norm)
        return self._max_velocity_norm

    @property
    def energy(self) -> np.ndarray:
        if self._energy is None:
            self._energy = utils.energy_from_velocity(self.velocity_norm)
        return self._energy

    @property
    def mean_energy(self) -> float:
        if self._mean_energy is None:
            self._mean_energy = np.float64(np.mean(self.energy))
        return self._mean_energy

    @property
    def std_energy(self) -> float:
        if self._std_energy is None:
            self._std_energy = np.float64(np.std(self.energy))
        return self._std_energy

    @property
    def max_energy(self) -> float:
        if self._max_energy is None:
            self._max_energy = np.max(self.energy)
        return self._max_energy

    @property
    def acceleration_norm(self) -> np.ndarray:
        if self._acceleration_norm is None:
            self._acceleration_norm = linalg.norm(self.acceleration, axis=0)
        return self._acceleration_norm

    @property
    def max_acceleration_norm(self) -> float:
        if self._max_acceleration_norm is None:
            self._max_acceleration_norm = np.max(self.acceleration_norm)
        return self._max_acceleration_norm

    @property
    def energy_distribution(self) -> InstantaneousEnergyDistribution:
        if self._energy_distribution.eedf is None:
            self._energy_distribution.calculate_distribution(self.energy)
        return self._energy_distribution

    def reset_cache(self) -> None:
        """
        Resets the cache to ensure it is updated.
        """

        self._mean_position = None
        self._var_position = None
        self._mean_velocity = None
        self._mean_velocity_moment = None
        self._energy = None
        self._mean_energy = None
        self._std_energy = None
        self._max_energy = None
        self._position_norm = None
        self._velocity_norm = None
        self._max_velocity_norm = None
        self._acceleration_norm = None
        self._max_acceleration_norm = None
        self._energy_distribution = InstantaneousEnergyDistribution()

    def plot_all(self, show: bool = True, block: bool = True) -> None:
        """
        Produces three Matplotlib Figures to visualize the position, velocity, and
        energy distribution of the electrons.

        Args:
            show (bool): calls plt.show() if True, else does nothing
            block (bool): if show is True, plt.show(block) is called
        """

        self.plot_position(show=False, block=False)
        self.plot_velocity(show=False, block=False)
        self.plot_energy(show=show, block=block)

    def plot_position(self, show: bool = True, block: bool = True) -> plt.Figure:
        """
        Produces a figure showing the electron positions. The figure contains four
        subplots showing the 3D scatter, and histograms along x, y, and z directions of
        the electron positions.

        Args:
            show (bool): calls plt.show() if True, else does nothing
            block (bool): if show is True, plt.show(block) is called
        Returns: Matplotlib figure object
        """

        fig = plt.figure()
        plt.suptitle("Electron positions")
        ax = fig.add_subplot(221, projection='3d')
        ax.scatter(self.position[0, :], self.position[1, :], self.position[2, :])
        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')
        ax.set_zlabel('z (m)')
        ax = fig.add_subplot(222)
        (_, bins) = np.histogram(self.position[0, :], bins='auto')
        ax.hist(self.position[0, :], bins=bins)
        ax.set_xlabel('x (m)')
        ax.set_ylabel('number of electrons')
        ax = fig.add_subplot(223)
        (_, bins) = np.histogram(self.position[1, :], bins='auto')
        ax.hist(self.position[1, :], bins=bins)
        ax.set_xlabel('y (m)')
        ax.set_ylabel('number of electrons')
        ax = fig.add_subplot(224)
        (_, bins) = np.histogram(self.position[2, :], bins='auto')
        ax.hist(self.position[2, :], bins=bins)
        ax.set_xlabel('z (m)')
        ax.set_ylabel('number of electrons')

        if show:
            plt.show(block=block)
        return fig

    def plot_velocity(self, show: bool = True, block: bool = True) -> plt.Figure:
        """
        Produces a figure showing the electron positions in velocity space. The figure
        contains four subplots showing the 3D scatter, and histograms along v_x, v_y,
        and v_z directions.

        Args:
            show (bool): calls plt.show() if True, else does nothing
            block (bool): if show is True, plt.show(block) is called
        Returns: Matplotlib figure object
        """

        fig = plt.figure()
        plt.suptitle("Electron velocities")
        ax = fig.add_subplot(221, projection='3d')
        ax.scatter(self.velocity[0, :], self.velocity[1, :], self.velocity[2, :])
        ax.set_xlabel('v$_x$ (m s$^{-1}$)')
        ax.set_ylabel('v$_y$ (m s$^{-1}$)')
        ax.set_zlabel('v$_z$ (m s$^{-1}$)')
        ax = fig.add_subplot(222)
        (_, bins) = np.histogram(self.velocity[0, :], bins='auto')
        ax.hist(self.velocity[0, :], bins=bins)
        ax.set_xlabel('v$_x$ (m s$^{-1}$)')
        ax.set_ylabel('number of electrons')
        ax = fig.add_subplot(223)
        (_, bins) = np.histogram(self.velocity[1, :], bins='auto')
        ax.hist(self.velocity[1, :], bins=bins)
        ax.set_xlabel('v$_y$ (m s$^{-1}$)')
        ax.set_ylabel('number of electrons')
        ax = fig.add_subplot(224)
        (_, bins) = np.histogram(self.velocity[2, :], bins='auto')
        ax.hist(self.velocity[2, :], bins=bins)
        ax.set_xlabel('v$_z$ (m s$^{-1}$)')
        ax.set_ylabel('number of electrons')

        if show:
            plt.show(block=block)
        return fig

    def plot_energy(self, show: bool = True, block: bool = True) -> plt.Figure:
        """
        Produces a figure showing the electron energy distribution function and the
        electron energy probability function.

        Args:
            show (bool): calls plt.show() if True, else does nothing
            block (bool): if show is True, plt.show(block) is called
        Returns: Matplotlib figure object
        """

        distri = self.energy_distribution

        fig, ax1 = plt.subplots()
        plt.title("Electron Energy Distribution Function (EEDF) \n"
                  "and Electron Energy Probability Function (EEPF)")
        color = 'tab:blue'
        ax1.hist(distri.energy_bin_centers, bins=distri.energy_bins,
                 weights=distri.eedf, color=color, histtype='step')
        ax1.tick_params(axis='y', labelcolor=color)
        ax1.set_ylabel('EEDF (eV$^{-1}$)', color=color)
        ax1.set_xlabel('energy (eV)')
        ax2 = ax1.twinx()
        color = 'tab:red'
        ax2.hist(distri.energy_bin_centers, bins=distri.energy_bins,
                 weights=distri.eepf, color=color, histtype='step')
        ax2.tick_params(axis='y', labelcolor=color)
        ax2.set_ylabel('EEPF (eV$^{-3/2}$)', color=color)
        fig.tight_layout()

        if show:
            plt.show(block=block)
        return fig



#  Copyright (c) 2020 ETH Zurich

from .simulation import Simulation  # noqa: F401
from .monte_carlo import MonteCarlo  # noqa: F401
from .electrons import Electrons  # noqa: F401
from .output import Output  # noqa: F401
from .config import Config  # noqa: F401
from .gas_mixture import GasMixture  # noqa: F401

from .__about__ import (  # noqa: F401
    __author__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __url__,
    __version__,
)

__all__ = [
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__url__",
    "__version__",
]



#  Copyright (c) 2020 ETH Zurich

"""
Module for the Simulation class.
"""

# Import Packages
import pickle
import numpy as np
import os
from typing import Union

# Import modules
from pyMETHES.__about__ import __version__
from pyMETHES.config import Config
from pyMETHES.monte_carlo import MonteCarlo
from pyMETHES.gas_mixture import GasMixture
from pyMETHES.electrons import Electrons
from pyMETHES.output import Output


class Simulation:
    """
    Main class of the pyMETHES simulation tool.

    The Simulation can be initialized providing the path to a configuration file, or
    a configuration dictionary. A different configuration can be applied at later stages
    using the apply_config method. A single simulation can be run with the run method,
    a series of simulation with the run_series method.

    Attributes:
        config (Config): configuration of the simulation
        mc (MonteCarlo): Monte-Carlo methods
        gas_mixture (GasMixture): GasMixture object containing the cross section data
        electric_field (float): electric field strength (V.m-1)
        electrons (Electrons): electron related data
        output (Output): output data of the simulation
    """

    version = f"pyMETHES version {__version__}\n"

    def __init__(self, config: Union[str, dict]):
        """
        Instantiates a Simulation.

        Args:
            config (str, dict): path to a json or json5 config file, or dictionary.
        """

        self.config = None
        self.mc = None
        self.gas_mixture = None
        self.electric_field = None
        self.electrons = None
        self.output = None
        self.apply_config(config)

    def apply_config(self, config: Union[str, dict] = None) -> None:
        """
        Applies the config to (re-)initialize all attributes of Simulation.

        Args:
            config (str, dict): path to a json or json5 config file, or dictionary
                containing the configuration for the simulation
        """

        if config is not None:
            self.config = Config(config)

        # creates the output directory if it does not already exist
        if not os.path.isdir(self.config.output_directory):
            os.mkdir(self.config.output_directory)

        self.mc = MonteCarlo(self.config)

        self.gas_mixture = GasMixture(self.config.gases,
                                      self.config.paths_to_cross_section_files,
                                      self.config.fractions,
                                      self.config.max_cross_section_energy)

        # Electric field strength in the z direction
        self.electric_field = self.config.EN * 1e-21 * self.config.gas_number_density

        self.electrons = Electrons(self.config.num_e_initial,
                                   self.config.initial_pos_electrons,
                                   self.config.initial_std_electrons,
                                   self.electric_field)

        self.output = Output(self.config, self.version, self.electrons)

    def run(self) -> None:
        """
        Runs the simulation, until one of the end conditions is fulfilled.
        """

        self.mc.calculate_max_coll_freq(self.gas_mixture)
        while not self.end_simulation():
            self.advance_one_step()
            self.print_step_info()
        self.calculate_final_output()
        self.save()

    def run_series(self, param: str, values: np.ndarray) -> None:
        """
        Runs a series of simulations.

        Args:
            param (str): name of the configuration parameter to be varied
            values (ndarray): list of values of the configuration parameter for the
                different simulations
        """

        base_name = self.config.base_name
        cfg_dict = self.config.to_dict()

        category = [cat for cat in cfg_dict.keys() if param in cfg_dict[cat].keys()]
        if not category:
            raise ValueError(f"{param} is not a valid configuration parameter name.")
        else:
            category = category[0]

        digits = len(str(values.size))
        for i, val in enumerate(values):
            cfg_dict['output']['base_name'] = base_name + f"_{i:0{digits}}"
            cfg_dict[category][param] = val
            self.apply_config(cfg_dict)
            self.run()

    def save(self) -> None:
        """
        Save the data specified in the configuration: pickle file of the simulation,
        temporal evolution of the swarm, swarm parameters, electron energy distribution.
        """

        if self.config.save_simulation_pickle:
            self.save_pickle()
        if self.config.save_temporal_evolution:
            self.output.save_temporal_evolution()
        if self.config.save_swarm_parameters:
            self.output.save_swarm_parameters()
        if self.config.save_energy_distribution:
            self.output.save_energy_distribution()

    def save_pickle(self, name: str = None) -> None:
        """
        Save the MonteCarlo class instance to a pickle file.

        Args:
            name: name of the pickle file created
        """

        if name is None:
            name = '_'.join([self.config.base_name, "simulation"])

        with open(self.config.output_directory + name + ".pickle", "wb") as pickle_file:
            pickle.dump(self, pickle_file)

    def advance_one_step(self) -> None:
        """
        Advances the simulation by one time step.
        """

        dt = self.mc.determine_timestep(self.electrons.max_velocity_norm,
                                        self.electrons.max_acceleration_norm)
        self.mc.determine_collisions(self.gas_mixture, self.electrons.velocity_norm,
                                     self.electrons.energy)
        pos, vel, nc, ni, na = self.mc.perform_collisions(self.gas_mixture,
                                                          self.electrons.position,
                                                          self.electrons.velocity,
                                                          self.electrons.energy)
        self.electrons.apply_scatter(pos, vel, self.electric_field)
        self.electrons.free_flight(dt)
        self.collect_output_data(dt, nc, ni, na)

    def collect_output_data(self, dt, nc, ni, na):
        """
        Collects and store current data for the simulation output.

        Args:
            dt (float): duration of the current time-step (s)
            nc (int): number of collisions during the current time-step
            ni (int): number of cations produced during the current time-step
            na (int): number of anions produced during the current time-step
        """

        self.output.time_series.append_data(self.electrons, dt, nc, ni, na)
        if self.output.time_series.ind_equ is None:
            if self.output.check_sst():
                # when equilibrium is reached, generate fixed energy bins to start
                # averaging the eedf over time:
                self.output.energy_distribution.generate_bins(
                    self.config.num_energy_bins,
                    self.electrons.max_energy
                )
        else:
            # updates the time-averaged energy distribution of electrons
            self.output.energy_distribution.collect_histogram(
                self.electrons.energy)
            # calculates the flux data (end condition for the simulation)
            self.output.flux.calculate_data(self.output.time_series)

    def calculate_final_output(self) -> None:
        """
        Calculates the final output data at the end of the simulation.
        """

        if self.output.time_series.ind_equ is not None:
            self.output.flux.calculate_data(self.output.time_series)
            self.output.bulk.calculate_data(self.output.time_series)
            self.output.energy_distribution.calculate_distribution(
                self.output.time_series.mean_energy[self.output.time_series.ind_equ:])
            self.output.rates_conv.calculate_data(self.gas_mixture,
                                                  self.output.energy_distribution)
            self.output.rates_count.calculate_data(self.output.time_series)

    def print_step_info(self) -> None:
        """
        Prints information on the current simulation step: mean electron energy,
        relative error of the flux drift velocity in z direction, and relative error
        of the flux diffusion coefficient (maximum of x, y, z directions).
        """

        info = f"Mean energy: {self.electrons.mean_energy:6.2f} eV."
        try:
            rel_err_w = self.output.flux.w_err[2] / self.output.flux.w[2]
            rel_err_d = np.max(self.output.flux.DN_err / self.output.flux.DN)
            info += (f" Error of w: {100 * rel_err_w:5.2f}%."
                     f" Error of DN: {100 * rel_err_d:5.2f}%.")
        except FloatingPointError:
            pass
        print(info)

    def end_simulation(self) -> bool:
        """
        Check end conditions for the simulation.

        Returns: True if simulation is finished, False otherwise.
        """

        end = False
        if not self.config.conserve:
            if self.electrons.num_e <= 0:
                end = True
                print('Simulation ended: number of electrons is zero')
            if self.electrons.num_e >= self.config.num_e_max:
                self.config.conserve = True
                print('Maximum number of electrons reached. Conserving the number '
                      'of electrons for the rest of the simulation.')

        # condition on the convergence of flux w and flux DN
        if not np.isnan(self.output.flux.DN_err[2]):
            w = self.output.flux.w
            w_err = self.output.flux.w_err
            DN = self.output.flux.DN
            DN_err = self.output.flux.DN_err
            if w_err[2] <= w[2] * self.config.w_tol \
                    and all(DN_err <= DN * self.config.DN_tol):
                end = True
                print("Simulation ended."
                      f" Error of w < {100 * self.config.w_tol:5.3}%."
                      f" Error of DN < {100 * self.config.DN_tol:5.3}%.")

        if self.output.time_series.num_collisions[-1] >= self.config.num_col_max:
            end = True
            print('Simulation ended: maximum number of collisions reached')

        return end



#  Copyright (c) 2020 ETH Zurich

import io
import os
import re

from setuptools import find_packages
from setuptools import setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())


about = {}
with open("pyMETHES/__about__.py") as fp:
    exec(fp.read(), about)

requirements = [
    'json5', 'lxcat_data_parser', 'matplotlib', 'scipy', 'numpy', 'pandas', 'molmass']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    name=about["__title__"],
    version=about["__version__"],
    url=about["__url__"],
    license=about["__license__"],
    author=about["__author__"],
    author_email=about["__email__"],
    description=about["__summary__"],

    long_description=read("README.rst"),

    include_package_data=True,

    packages=find_packages(exclude=('tests',)),

    install_requires=requirements,
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Physics',
    ],
)



#  Copyright (c) 2020 ETH Zurich

"""
Module with some utility methods.
"""

import numpy as np
import scipy.constants as sc
from typing import Union

num = Union[int, float, np.ndarray]


def velocity_from_energy(energy: num) -> num:
    """
    Calculate the velocity norm of electrons, based on their kinetic energy.

    Args:
        energy: Energy of the electron (eV)

    Returns: Velocity norm of the electrons (m.s-1)
    """

    return np.sqrt(2 * energy * sc.elementary_charge / sc.electron_mass)


def energy_from_velocity(velocity: num) -> num:
    """
    Calculate the kinetic energy of electrons, based on the norm of their velocity.

    Args:
        velocity: velocity norm of the electrons (m.s-1)

    Returns: Energy of the electrons (eV)
    """

    return 0.5 * sc.electron_mass * velocity ** 2 / sc.elementary_charge


def acceleration_from_electric_field(electric_field: num) -> num:
    """
    Calculates the acceleration of electrons, based on the local electric field.

    Args:
        electric_field: local electric field strength (V.m-1)

    Returns: acceleration (m.s-2)
    """

    return electric_field * (sc.elementary_charge / sc.electron_mass)



#  Copyright (c) 2020 ETH Zurich

"""
Module for the Output class.
"""

# Import Packages
import numpy as np
import pandas as pd
import json
import matplotlib.pyplot as plt

# Import modules
from pyMETHES.config import Config
from pyMETHES.electrons import Electrons
from pyMETHES.energy_distribution import TimeAveragedEnergyDistribution
from pyMETHES.transport_data import BulkData, FluxData
from pyMETHES.rate_coefficients import ConvolutedRates, CountedRates
from pyMETHES.temporal_evolution import TimeSeries

np.seterr(all='raise')


class Output:
    """
    The Output class instantiates all output-related classes: TimeSeries,
    TimeAveragedEnergyDistribution, FluxData, BulkData, ConvolutedRates, CountedRates.
    It also provides methods to save or to plot the output data. Finally, it provides
    the check_sst method which test if the swarm is at equilibrium based on the
    evolution of the mean electron energy.

    Attributes:
        config (Config): configuration of the simulation
        version (str): version of pyMETHES as a string
        time_series (TimeSeries): temporal evolution data
        energy_distribution (TimeAveragedEnergyDistribution): energy distribution data
        flux (FluxData): flux transport data
        bulk (BulkData): flux transport data
        rates_conv (ConvolutedRates): convoluted rate coefficients
        rates_count (CountedRates): counted rate coefficients
    """

    def __init__(self, cfg: Config, ver: str, electrons: Electrons):
        """
        Instantiates the Output class.

        Args:
            cfg (Config): configuration of the simulation
            ver (str): version of pyMETHES as a string
            electrons (Electrons): electron data
        """

        self.config = cfg
        self.version = ver

        # temporal evolution of some quantities
        self.time_series = TimeSeries(electrons)

        # energy-related data:
        self.energy_distribution = TimeAveragedEnergyDistribution()

        # flux data
        self.flux = FluxData(self.config.gas_number_density)

        # bulk data
        self.bulk = BulkData(self.config.gas_number_density)

        # rate coefficients calculated by convolution
        self.rates_conv = ConvolutedRates()

        # rate coefficients calculated by counting events
        self.rates_count = CountedRates(cfg.gas_number_density, cfg.conserve)

    def check_sst(self) -> bool:
        """
        Checks if the swarm energy is at equilibrium (steady-state). This is done by
        checking if the mean electron energy has dropped during the last 10% of the
        elapsed time, compared to the period between 80 and 90%.

        Returns: True if equilibrium was reached, False otherwise
        """

        # Intervals need to contain at least 10 elements
        n = round(self.time_series.mean_energy.size / 10)
        # Checks if the interval 80-90 % of energy data is larger
        # than the interval 90-100%
        if n >= 10 and np.mean(self.time_series.mean_energy[-2*n:-n]) >= \
                np.mean(self.time_series.mean_energy[-n:]):
            # store the index, where the sst-data starts
            self.time_series.ind_equ = self.time_series.time.size
            return True
        return False

    def save_temporal_evolution(self, name: str = None) -> None:
        """
        Saves the temporal evolution of the swarm data to a json file.

        Args:
            name: name of the json file created
        """

        if name is None:
            name = '_'.join([self.config.base_name, "temporal_evolution"])

        with open(self.config.output_directory + name + ".json", "w") as output_file:
            json.dump(self.time_series.to_dataframe().to_dict('list'), output_file,
                      indent=2)

    def save_swarm_parameters(self, name: str = None) -> None:
        """
        Saves the final swarm parameters to a json file.

        Args:
            name: name of the json file created
        """

        if self.flux.w_err[2] is not None:

            coefficients = {
                'E/N (Td)': float(self.config.EN),
                'mean energy (eV)': self.energy_distribution.energy_mean,
                'mean energy error (eV)': self.energy_distribution.energy_mean_err,
                'bulk drift velocity (m.s-1)': self.bulk.w[2],
                'bulk drift velocity error (m.s-1)': self.bulk.w_err[2],
                'bulk L diffusion coeff. * N (m-1.s-1)': self.bulk.DN[2],
                'bulk L diffusion coeff. error * N (m-1.s-1)':
                    self.bulk.DN_err[2],
                'bulk T diffusion coeff. * N (m-1.s-1)': self.bulk.DN[0],
                'bulk T diffusion coeff. error * N (m-1.s-1)':
                    self.bulk.DN_err[0],
                'flux drift velocity (m.s-1)': self.flux.w[2],
                'flux drift velocity error (m.s-1)': self.flux.w_err[2],
                'flux L diffusion coeff. * N (m-1.s-1)': self.flux.DN[2],
                'flux L diffusion coeff. error * N (m-1.s-1)':
                    self.flux.DN_err[2],
                'flux T diffusion coeff. * N (m-1.s-1)': self.flux.DN[0],
                'flux T diffusion coeff. error * N (m-1.s-1)':
                    self.flux.DN_err[0],
                'effective ionization rate coeff. (counted) (m3.s-1)':
                    self.rates_count.effective,
                'effective ionization rate coeff. (counted) error (m3.s-1)':
                    self.rates_count.effective_err,
                'ionization rate coeff. (counted) (m3.s-1)':
                    self.rates_count.ionization,
                'ionization rate coeff. (counted) error (m3.s-1)':
                    self.rates_count.ionization_err,
                'attachment rate coeff. (counted) (m3.s-1)':
                    self.rates_count.attachment,
                'attachment rate coeff. (counted) error (m3.s-1)':
                    self.rates_count.attachment_err,
                'effective ionization rate coeff. (convolution) (m3.s-1)':
                    self.rates_conv.effective,
                'ionization rate coeff. (convolution) (m3.s-1)':
                    self.rates_conv.ionization,
                'attachment rate coeff. (convolution) (m3.s-1)':
                    self.rates_conv.attachment,
            }

            if name is None:
                name = '_'.join([self.config.base_name, "swarm_parameters"])

            with open(self.config.output_directory
                      + name + ".json", "w") as output_file:
                json.dump(coefficients, output_file, indent=2)

    def save_energy_distribution(self, name: str = None) -> None:
        """
        Saves the final time-averaged electron energy distribution function and electron
        energy probability function to a json file.

        Args:
            name: name of the json file created
        """

        if self.energy_distribution.eedf is not None:
            energy_distributions = pd.DataFrame({
                'energy (eV)': self.energy_distribution.energy_bin_centers,
                'eedf (eV-1)': self.energy_distribution.eedf,
                'eepf (eV-3/2)': self.energy_distribution.eepf
            }).to_dict('list')

            if name is None:
                name = '_'.join([self.config.base_name, "energy_distribution"])

            with open(self.config.output_directory
                      + name + ".json", "w") as output_file:
                json.dump(energy_distributions, output_file, indent=2)

    @staticmethod
    def plot_sst_line(axes, t_sst) -> None:
        """
        Plots a vertical line at the time instant where equilibrium is reached.

        Args:
            axes: axes to plot the line on
            t_sst: equilibration time
        """

        y_limits = axes.get_ylim()
        axes.plot([t_sst, t_sst], y_limits, 'k-', label='equilibration time')
        axes.set_ylim(y_limits)

    def plot_temporal_evolution(self, show: bool = True,
                                block: bool = True) -> plt.Figure:
        """
        Produces a figure showing the temporal evolution of the swarm. The figure
        contains five subplots showing the number of particles, the mean electron
        energy, the number of collisions, the mean electron position and the
        variance of electron positions.

        Args:
            show (bool): calls plt.show() if True, else does nothing
            block (bool): if show is True, plt.show(block) is called

        Returns: Matplotlib figure object
        """

        data = self.time_series
        simu_time = data.time
        pos_mean = data.mean_position
        pos_variance = data.var_position

        # plot a line to indicate the time when equilibrium is reached
        t_sst = None
        if self.time_series.ind_equ is not None:
            t_sst = data.time[self.time_series.ind_equ]

        fig = plt.figure()
        ax = fig.add_subplot(231)
        ax.plot(simu_time, data.num_electrons - data.num_electrons[0],
                label='n_electrons - n_0', linewidth=3)
        ax.plot(simu_time, data.num_cations, label='n_cations')
        ax.plot(simu_time, data.num_anions, label='n_anions')
        Output.plot_sst_line(ax, t_sst)
        ax.legend()
        ax.set_title('Number of Particles')
        ax.set_ylabel('number of particles')
        ax.set_xlabel('time (s)')
        ax.grid(True)

        ax = fig.add_subplot(232)
        ax.plot(simu_time, data.mean_energy)
        Output.plot_sst_line(ax, t_sst)
        ax.legend()
        ax.set_title("Mean Energy of Electrons")
        ax.set_ylabel('mean electron energy (eV)')
        ax.set_xlabel('time (s)')
        ax.grid(True)

        ax = fig.add_subplot(233)
        ax.plot(simu_time[1:], np.diff(data.num_collisions), label='collisions')
        ax.plot(simu_time[1:], data.num_electrons[1:] - np.diff(data.num_collisions),
                label='null collisions')
        Output.plot_sst_line(ax, t_sst)
        ax.legend()
        ax.set_title('Number of Collisions Per Simulation Step')
        ax.set_ylabel('number of collisions')
        ax.set_xlabel('time (s)')
        ax.grid(True)

        ax = fig.add_subplot(234)
        ax.plot(simu_time, pos_mean[:, 0], label="x")
        ax.plot(simu_time, pos_mean[:, 1], label="y")
        ax.plot(simu_time, pos_mean[:, 2], label="z")
        Output.plot_sst_line(ax, t_sst)
        ax.legend()
        ax.set_title("Mean Position of Electrons")
        ax.set_ylabel('mean position (m)')
        ax.set_xlabel('time (s)')
        ax.grid(True)

        ax = fig.add_subplot(235)
        ax.plot(simu_time, pos_variance[:, 0], label="x")
        ax.plot(simu_time, pos_variance[:, 1], label="y")
        ax.plot(simu_time, pos_variance[:, 2], label="z")
        Output.plot_sst_line(ax, t_sst)
        ax.legend()
        ax.set_title("Variance of Electrons Positions")
        ax.set_ylabel('variance (m$^2$)')
        ax.set_xlabel('time (s)')
        ax.grid(True)

        if show:
            plt.show(block=block)
        return fig

    def plot_energy_distribution(self, show: bool = True,
                                 block: bool = True) -> plt.Figure:
        """
        Produces a figure showing the time-averaged electron energy distribution
        function and electron energy probability function.

        Args:
            show (bool): calls plt.show() if True, else does nothing
            block (bool): if show is True, plt.show(block) is called

        Returns: Matplotlib figure object
        """

        if self.energy_distribution.eedf is not None:
            energy = self.energy_distribution.energy_bin_centers

            fig, ax1 = plt.subplots()
            plt.title("Electron Energy Distribution Function (EEDF) \n"
                      "and Electron Energy Probability Function (EEPF)")
            color = 'tab:blue'
            ax1.plot(energy, self.energy_distribution.eedf, color=color)
            ax1.tick_params(axis='y', labelcolor=color)
            ax1.set_ylabel('EEDF (eV$^{-1}$)', color=color)
            ax1.set_xlabel('energy (eV)')
            ax1.grid(True)

            ax2 = ax1.twinx()
            color = 'tab:red'
            ax2.plot(energy, self.energy_distribution.eepf, color=color)
            ax2.tick_params(axis='y', labelcolor=color)
            ax2.set_ylabel('EEPF (eV$^{-3/2}$)', color=color)

            if show:
                plt.show(block=block)
            return fig



#  Copyright (c) 2020 ETH Zurich

"""
Module for the CountedRates and ConvolutedRates classes.

Both classes are based on the RateCoefficients abstract class. They implement two
different methods for calculating ionization/attachment rate coefficients. The
CountedRates class calculates the rate coefficients by averaging over time the number
of ionization and attachment collisions. The ConvolutedRates class calculates the
rate coefficients by calculating the convolution product of the cross section with the
time-averaged electron energy distribution of the electrons.
"""

# Import Packages
from abc import ABC
import numpy as np
import scipy.constants as csts

# Import modules
from pyMETHES.energy_distribution import TimeAveragedEnergyDistribution
from pyMETHES.gas_mixture import GasMixture
from pyMETHES.temporal_evolution import TimeSeries

np.seterr(all='raise')


class RateCoefficients(ABC):
    """
    Ionization, attachment and effective ionization rate coefficients

    Attributes:
        ionization: ionization reaction rate coefficient (m3.s-1)
        attachment: attachment reaction rate coefficient (m3.s-1)
        effective: effective ionization reaction rate coefficient (m3.s-1)
    """

    def __init__(self):

        self.ionization = None
        self.attachment = None
        self.effective = None


class ConvolutedRates(RateCoefficients):
    """
    Ionization, attachment and effective ionization rate coefficients calculated by
    convolution product of eedf and cross section.
    """

    def calculate_data(self, mix: GasMixture,
                       distri: TimeAveragedEnergyDistribution) -> None:
        """
        Calculates the effective, ionization and attachment rate coefficients.

        Args:
            mix (GasMixture): GasMixture object containing the cross section data
            distri (TimeAveragedEnergyDistribution): electron energy distribution (eV-1)
        """

        self.ionization = np.sum([self.calculate_convolution(x, distri)
                                  for x in mix.cross_sections[mix.is_ionization]])
        self.attachment = np.sum([self.calculate_convolution(x, distri)
                                  for x in mix.cross_sections[mix.is_attachment]])
        self.effective = self.ionization - self.attachment

    @staticmethod
    def calculate_convolution(interp,
                              distri: TimeAveragedEnergyDistribution) -> float:
        """
        Calculates the reaction rate of a collision type by calculating the convolution
        product of the eedf and the corresponding cross sections.

        Args:
            interp: linear interpolation of the cross section (m2)
            distri (TimeAveragedEnergyDistribution): energy distribution of the
                electrons (eV-1)

        Returns: reaction rate coefficient (m3.s-1)
        """

        cross_section = interp(distri.energy_bin_centers)

        A = distri.eedf * distri.energy_bin_centers ** 0.5 * np.diff(distri.energy_bins)
        rrate = np.sqrt(
            2 * csts.elementary_charge / csts.electron_mass) * np.dot(A, cross_section)
        return rrate


class CountedRates(RateCoefficients):
    """
    Ionization, attachment and effective ionization rate coefficients, calculated by
    counting ionization and attachment collisions.

    Attributes:
        gas_number_density (float): gas number density (m-3)
        conserve (bool): conservation of the electron number
    """

    def __init__(self, gas_number_density: float, conserve: bool):
        """
        Instantiation of CountedRates.

        Args:
            gas_number_density (float): gas number density (m-3) for normalization
            conserve (bool): bool indicating conservation of the electron number
                (affects the statistics when counting the collisions)
        """

        super().__init__()

        self.gas_number_density = gas_number_density
        self.conserve = conserve
        self.ionization_err = None
        self.attachment_err = None
        self.effective_err = None

    def calculate_data(self, time_series: TimeSeries) -> None:
        """
        Calculates the ionization, attachment and effective ionization rate coefficients
        values by counting collision events.

        Args:
            time_series (TimeSeries): temporal evolution of some quantities
        """

        steps_since_sst = time_series.time.size - 1 - time_series.ind_equ
        x = time_series.time[time_series.ind_equ:] \
            - time_series.time[time_series.ind_equ]

        if not self.conserve:
            # Ne(t-t_SST) = Ne(t_SST) * exp(nu_eff * (t-t_SST))
            # log(Ne(t-t_SST)) - log(Ne(t_SST)) = nu_eff * (t-t_SST)
            y = np.log(time_series.num_electrons[time_series.ind_equ:]) \
                - np.log(time_series.num_electrons[time_series.ind_equ])
            d = y[1:] / x[1:]
            nu_eff = np.mean(d)
            if nu_eff != 0:
                x = (np.exp(nu_eff * x) - 1) / nu_eff \
                    * time_series.num_electrons[time_series.ind_equ]
        else:
            y = time_series.num_electrons[time_series.ind_equ:] \
                - time_series.num_electrons[time_series.ind_equ]
            d = y[1:] / x[1:]

        # effective ionization
        self.effective = np.mean(d) / self.gas_number_density
        self.effective_err = np.std(d) / np.sqrt(steps_since_sst) \
            / self.gas_number_density

        # ionization (cations, positively charged)
        y = time_series.num_cations[time_series.ind_equ:] \
            - time_series.num_cations[time_series.ind_equ]
        d = y[1:] / x[1:]
        self.ionization = np.mean(d) / self.gas_number_density
        self.ionization_err = \
            np.std(d) / np.sqrt(steps_since_sst) / self.gas_number_density

        # attachment (anions, negatively charged)
        y = time_series.num_anions[time_series.ind_equ:] \
            - time_series.num_anions[time_series.ind_equ]
        d = y[1:] / x[1:]
        self.attachment = np.mean(d) / self.gas_number_density
        self.attachment_err = \
            np.std(d) / np.sqrt(steps_since_sst) / self.gas_number_density
